import pdb
import torch
from numpy.polynomial.legendre import leggauss

def leggaussVarLimits(n, a, b, dtype=torch.double):
    """
    Computers n weights and points for Gauss-Legendre numerical integration of a
    function in the interval (a,b).
    """
    x, w = leggauss(deg=n)
    # reversing x and w to make them compatible with Lea's Matlab legquad function
    # x = x[::-1].copy()
    # w = w[::-1].copy()
    x = torch.from_numpy(x)
    w = torch.from_numpy(w)
    xVarLimits = (x*(b-a)+(b+a))/2
    wVarLimits = (b-a)/2*w
    return xVarLimits, wVarLimits




class FileLock:
    def __init__(self, filename):
        self.filename = filename
        with open(self.filename, 'w') as f:
            f.write('done')

    def lock(self):
        with open(self.filename, 'w') as f:
            f.write('working')

    def unlock(self):
        with open(self.filename, 'w') as f:
            f.write('done')

    def is_locked(self):
        return open(self.filename, 'r').read() == 'working'





import pdb
import math
import torch
import scipy.stats

class GaussianProcess(object):

    def __init__(self, mean, kernel):
        self._mean = mean
        self._kernel = kernel

    def __call__(self, t, epsilon=1e-5):
        return self.eval(t=t, epsilon=epsilon)

    def eval(self, t, regularization=1e-5):
        mean = self._mean(t)
        cov = self._kernel.buildKernelMatrix(t)
        cov = cov + regularization*torch.eye(cov.shape[0])
        mn = scipy.stats.multivariate_normal(mean=mean, cov=cov)
        samples = torch.from_numpy(mn.rvs())
        return samples, mean, cov

    def mean(self, t):
        return self._mean(t)

    def std(self, t):
        Kdiag = self._kernel.buildKernelMatrixDiag(t)
        std = torch.sqrt(Kdiag)
        return std




''' Dumps a config file of the type readable by configparser
into a dictionary 
Ref: http://docs.python.org/library/configparser.html

Downloaded from git@gist.github.com:c5607843dd7174b8d6828185a09584e7.git
'''

import sys
import configparser

class GetDict:

    def __init__(self, config):
        self.config = config

    def get_dict(self):
        # config = configparser.ConfigParser()
        # config.read(self.config)

        sections_dict = {}

        # get all defaults
        defaults = self.config.defaults()
        temp_dict = {}
        for key in defaults.keys():
            temp_dict[key] = defaults[key]

        sections_dict['default'] = temp_dict

        # get sections and iterate over each
        sections = self.config.sections()

        for section in sections:
            options = self.config.options(section)
            temp_dict = {}
            for option in options:
                temp_dict[option] = self.config.get(section,option)

            sections_dict[section] = temp_dict

        return sections_dict

if __name__== '__main__':

    if len(sys.argv) == 1:
        print('Must provide the path to the config file as the argument')
        sys.exit(1)

    config = configparser.ConfigParser()
    config.read(sys.argv[1])
    getdict = GetDict(config)
    config_dict = getdict.get_dict()

    # print the entire dictionary
    # Trick from http://stackoverflow.com/a/3314411/59634
    import json
    print(json.dumps(config_dict, indent=2))




import pdb
import torch

def sampleInhomogeneousPP_thinning(CIF_times, CIF_values):
    """ Thining algorithm to sample from an inhomogeneous point process. Algorithm 2 from Yuanda Chen (2016). Thinning algorithms for simulating Point Prcesses.

    :param: CIF_times: times at which the CIF is evaluated
    :type: CIF_times: list like

    :param: CIF_values: values of the CIF evaluated at CIF_times
    :type: CIF_values: list like

    :return: (inhomogeneous, homogeneous): samples of the inhomogeneous and homogenous point process with CIF values CIF_values evaluated at CIF_times
    :rtype: tuple containing two lists
    """
    m = 0
    t = [0]
    s = [0]
    T = CIF_times.max()
    lambda_max = CIF_values.max()
    while s[m]<T:
        u = torch.rand(1)
        w = -torch.log(u)/lambda_max   # w~exponential(lambda_max)
        s.append(s[m]+w)               # {sm} homogeneous Poisson process
        D = torch.rand(1)
        CIF_index = (CIF_times-s[m+1]).abs().argmin()
        approxCIF_atSpike = CIF_values[CIF_index]
        if D<=approxCIF_atSpike/lambda_max: # accepting with probability
                                            # cifF(s[m+1])/lambda_max
            t.append(s[m+1].item())         # {tn} inhomogeneous Poisson
                                                          # process
        m += 1
    if t[-1]<=T:
        answer = {"inhomogeneous": t[1:], "homogeneous": s[1:]}
    else:
        answer = {"inhomogeneous": t[1:-1], "homogeneous": s[1:-1]}
    return answer

def sampleInhomogeneousPP_timeRescaling(CIF_times, CIF_values, dt=0.001):
    """ Time rescaling algorithm to sample from an inhomogeneous point
    process. Chapter 2 from Uri Eden's Point Process Notes.

    :param: CIF_times: times at which the CIF is evaluated
    :type: CIF_times: list like

    :param: CIF_values: values of the CIF evaluated at CIF_times
    :type: CIF_values: list like

    :param: dt: spike-time resolution.
    :type: dt: float

    :return: samples of the inhomogeneous point process with CIF values CIF_values evaluated at CIF_times
    :rtype: list
    """
    m = 0
    t = [0]
    s = [0]
    T = CIF_times.max()
    while s[-1]<T:
        u = torch.rand(1)
        z = -torch.log(u)   # z~exponential(1.0)
        t = s[-1]
        CIF_index = (CIF_times-t).abs().argmin()
        approxCIF_t = CIF_values[CIF_index]
        anInt = approxCIF_t*dt
        while anInt<z and t<=T:
            t += dt
            CIF_index = (CIF_times-t).abs().argmin()
            approxCIF_t = CIF_values[CIF_index]
            anInt += approxCIF_t*dt
        s.append(t)
    if s[-1]<=T:
        answer = s[1:]
    else:
        answer = s[1:-1]
    return answer



import numpy as np

def bootstrapMean(observations, nResamples):
    indices = np.arange(len(observations))
    bootstrapped_means = np.empty(nResamples)
    for i in range(nResamples):
        sample_indices = np.random.choice(indices, size=len(indices), replace=True)
        resampled_observations = observations[sample_indices]
        if remove_nan:
            resampled_observations = \
                resampled_observations[~np.isnan(resampled_observations)]
        bootstrapped_means[i] = np.means(resampled_observations)
    return bootstrapped_means

def bootstrapMedian(observations, nResamples):
    indices = np.arange(len(observations))
    bootstrapped_medians = np.empty(nResamples)
    for i in range(nResamples):
        sample_indices = np.random.choice(indices, size=len(indices), replace=True)
        resampled_observations = observations[sample_indices]
        bootstrapped_medians[i] = np.median(resampled_observations)
    return bootstrapped_medians

def estimatePercentileCI(alpha, bootstrapped_stats):
    sortedSamples = np.sort(bootstrapped_stats)
    N = len(bootstrapped_stats)
    ciLow = sortedSamples[int(alpha/2*N)]
    ciHigh = sortedSamples[int((1-alpha/2)*N)]
    ci = (ciLow, ciHigh)
#     if ciLow>0:
#         import pdb; pdb.set_trace()
    return ci




import sys
import os
from scipy.io import loadmat
import torch
sys.path.append("../src")
import numericalMethods.utils

def test_leggaussVarLimits():
    tol = 1e-6
    dataFilename = os.path.join(os.path.dirname(__file__), "data/legquad.mat")
    mat = loadmat(dataFilename)
    n = mat["n"][0][0]
    a = mat["a"][0][0]
    b = mat["b"][0][0]

    x = torch.flip(input=torch.from_numpy(mat["x"]).type(torch.DoubleTensor), dims=[1])
    w = torch.flip(input=torch.from_numpy(mat["w"]).type(torch.DoubleTensor), dims=[1])

    px, pw = numericalMethods.utils.leggaussVarLimits(n=n, a=a, b=b, dtype=torch.double)
    xDiff = torch.mean((x-px)**2)
    assert(xDiff<tol)
    wDiff = torch.mean((w-pw)**2)
    assert(wDiff<tol)

if __name__=="__main__":
    test_leggaussVarLimits()




import pdb
import math
import torch
import scipy.stats

class GPMarginalLogLikelihood(object):
    def __init__(self, x, y, kernel):
        self._x = x
        self._y = y
        self._kernel = kernel

        self._kyInv = None
        self._logdetKy = None
        self._lastParams = None

    def _computeKyInvAndLogDet(self, params):
       ky = self._kernel.buildKSample(x1=self._x, x2=self._x, params=params)
       self._kyInv = torch.linalg.inv(a=ky)
       (_, self._logdetKy) = torch.linalg.slogdet(a=ky)
       self._lastParams = params

    def evalWithoutInverse(self, params):
        # K_y^-1 * y = u
        # y = K_y * u
        # y.T * K_y^-1 * y = y.T * u

        n = len(self._y)
        ky = self._kernel.buildKSample(x1=self._x, x2=self._x, params=params)
        u = torch.linalg.solve(a=ky, b=self._y)
        answer = -0.5*torch.dot(a=self._y, b=u)
        (_, logdetKy) = torch.linalg.slogdet(a=ky)
        answer = answer - 0.5*logdetKy - n/2* math.log(2*math.pi)

        return answer

    def eval(self, params):
        self._computeKyInvAndLogDet(params=params)
        n = len(self._y)
        answer = -.5*torch.dot(a=torch.dot(a=self._y, b=self._kyInv), b=self._y)\
                 -.5*self._logdetKy\
                 -n/2*math.log(2*math.pi)
        return answer

    def evalGradient(self, params):
        if (self._lastParams is None) or (not torch.array_equal(params, self._lastParams)):
            self._computeKyInvAndLogDet(params=params)

        alpha = torch.dot(a=self._kyInv, b=self._y)
        kyGrad = self._kernel.buildKSampleGrad(x1=self._x, x2=self._x, params=params)
        # kyGradValues = list(kyGrad.values())
        # gradValues = torch.empty(len(params))
        grad = torch.empty(len(params))
        matrixConst = torch.outer(a=alpha, b=alpha)-self._kyInv
        for i in range(len(kyGrad)):
            grad[i] = .5*torch.trace(torch.matmul(matrixConst, kyGrad[i]))
        return grad

    def evalWithGradient(self, params):
        value = self.eval(params=params)
        grad = self.evalGradient(params=params)
        answer = {}
        answer['value'] = value
        answer['grad'] = grad
        return answer




import pdb
import math
import warnings
import random
import pandas as pd
import numpy as np
import torch

from . import sampling

def timeRescaling(Y, pk):
    n = int(torch.sum(Y).item())
    if n==0:
        return None
    rISIs = torch.zeros(n-1, dtype=torch.double)
    spikeindicies = Y.nonzero()
    for r in range(n-1):
        ind1 = spikeindicies[r].item()
        ind2 = spikeindicies[r+1].item()
        rISIs[r] = torch.sum(pk[ind1+1:ind2+1]).item()
    return rISIs

def timeRescalingForUnbinnedSpikes(spikesTimes, cifValues, t0, tf, dt):
    pk = cifValues*dt
    bins = torch.arange(t0-dt/2, tf+dt/2, dt)
    # start binning spikes using pandas
    cutRes, _ = pd.cut(spikesTimes, bins=bins, retbins=True)
    Y = torch.from_numpy(cutRes.value_counts().values)
    # end binning spikes using pandas
    indicesMoreThanOneSpikes = (Y>1).nonzero()
    if len(indicesMoreThanOneSpikes)>0:
        warnings.warn("Found more than one spike in {:d} bins".format(len(indicesMoreThanOneSpikes)))
        Y[indicesMoreThanOneSpikes] = 1.0
    zExp = timeRescaling(Y=Y, pk=pk)
    return zExp

def KSTestTimeRescalingNumericalCorrection(spikesTimes, cifTimes, cifValues, gamma):
    # from Haslinger et al., 2010, p. 2492,
    # Procedure for Numerical Correction

    def subtractECDFs(ecdf1X, ecdf1Y, ecdf2X, ecdf2Y):
        subX = torch.sort(torch.cat((ecdf1X, ecdf2X)))[0].numpy()
        ecdf1YI = np.interp(subX, ecdf1X.numpy(), ecdf1Y.numpy())
        ecdf2YI = np.interp(subX, ecdf2X.numpy(), ecdf2Y.numpy())
        subY = ecdf1YI-ecdf2YI
        # begin debug
        # plt.plot(subX, ecdf1YI)
        # plt.plot(subX, ecdf2YI)
        # plt.show()
        # pdb.set_trace()
        # end debug
        return torch.tensor(subX), torch.tensor(subY)

    t0 = cifTimes.min()
    tf = cifTimes.max()
    dt = cifTimes[1]-cifTimes[0]
    T = tf-t0
    print("Processing given ISIs")
    zExp = timeRescalingForUnbinnedSpikes(spikesTimes=spikesTimes, cifValues=cifValues, t0=t0, tf=tf, dt=dt)
    zSim = None
    for i in range(gamma):
        print("Processing iter {:d}/{:d}".format(i, gamma-1))
        simSpikesTimes = sampling.sampleInhomogeneousPP_timeRescaling(CIF_times=cifTimes, CIF_values=cifValues)
        zSimIter = timeRescalingForUnbinnedSpikes(spikesTimes=simSpikesTimes,
                                                  cifValues=cifValues,
                                                  t0=t0, tf=tf, dt=dt)
        if zSimIter is not None:
            if zSim is None:
                zSim = zSimIter
            else:
                zSim = torch.cat((zSim, zSimIter))
    expECDFx, _ = torch.sort(zExp)
    expECDFy = torch.linspace(0, 1, len(zExp))
    simECDFx, _ = torch.sort(zSim)
    simECDFy = torch.linspace(0, 1, len(zSim))
    diffECDFsX, diffECDFsY = subtractECDFs(ecdf1X=expECDFx, ecdf1Y=expECDFy, ecdf2X=simECDFx, ecdf2Y=simECDFy)

    Nexp = len(zExp)
    Nsim = len(zSim)
    cb = 1.36*math.sqrt((Nexp+Nsim)/(Nexp*Nsim))

    return diffECDFsX, diffECDFsY, expECDFx, expECDFy, simECDFx, simECDFy, cb

def timeRescalingAnalyticalCorrection(Y, pk, eps=1e-10):
    indicesMoreThanOneSpikes = (Y>1).nonzero()
    if len(indicesMoreThanOneSpikes)>0:
        warnings.warn("Found more than one spike in {:d} bins".format(len(indicesMoreThanOneSpikes)))
        Y[indicesMoreThanOneSpikes] = 1.0
    pk = torch.max(pk, torch.tensor([eps], dtype=torch.double))
    qk = -torch.log(1-pk)
    # make the rescaled times
    n = int(torch.sum(Y).item())
    rISIs = torch.zeros(n-1)
    spikeindicies = Y.nonzero()
    for r in range(n-1):
        ind1 = spikeindicies[r].item()
        ind2 = spikeindicies[r+1].item()
        aSum = torch.sum(qk[(ind1+1):ind2]).item()
        # delta = -(1.0/qk[ind2].item())*math.log(1-random.random()*(1-math.exp(-qk[ind2].item())))
        # total = total + qk[ind2].item()*delta
        delta = -math.log(1-random.random()*(1-math.exp(-qk[ind2].item())))
        rISIs[r] = aSum + delta
    return rISIs

def KSTestTimeRescalingAnalyticalCorrection(Y, pk, eps=1e-10):
    '''
    Y: binary sequence for the discrete time point process
    pk: event probability in each time been based on a conditional cif function

    xks: uniform CDF
    rst: time rescaled times
    cb: 95% confidence bounds
    n: number of spikes

    KS-test based on time-rescalling theorem =====================
    Use DT Correction for Time Rescaling Theorem - Haslinger, Pipa and Brown (2010)
    discrete time conditional probabilities "pk"  where 0<=pk<= 1
    '''
    rISIs = timeRescalingAnalyticalCorrection(Y=Y, pk=pk, eps=eps) # rescaled inter-spike intervals
    utRISIs = 1-torch.exp(-rISIs)              # uniform-transformed srescaled inter-spike intervals
    srISIs, _ = torch.sort(rISIs)                 # sorted rescaled inter-spike intervals
    utSRISIs = 1-torch.exp(-srISIs)              # uniform-transformed sorted rescaled inter-spike intervals
    n = len(rISIs)
    # dt = 1/(n-1)
    # xks = torch.linspace(0.5*dt, 1-0.5*dt, 1.0/dt)
    k = torch.linspace(start=1, end=n, steps=n)
    uCDF = (k-0.5)/n                            # uniform cummulative distribution function
    cb = 1.36/math.sqrt(n)                      # 95% confidence bounds for large sample sizes (see Brown et al., 2001)
    return utSRISIs, uCDF, cb, utRISIs

def KSTestTimeRescalingAnalyticalCorrectionUnbinned(spikesTimes, cifValues, t0, tf, dt, eps=1e-10):
        pk = cifValues*dt
        bins = torch.arange(t0-dt/2, tf+dt/2, dt)
        # start binning spikes using pandas
        cutRes, _ = pd.cut(spikesTimes, bins=bins, retbins=True)
        Y = torch.from_numpy(cutRes.value_counts().values)
        # end binning spikes using pandas
        utSRISIs, uCDF, cb, utRISIs = KSTestTimeRescalingAnalyticalCorrection(Y=Y, pk=pk, eps=eps)
        return utSRISIs, uCDF, cb, utRISIs




import sys
import numpy as np

sys.path.append("../src")
import utils.neuralDataAnalysis

def test_binSpikes():
    spikes_times = np.arange(0.0, 9.99, 0.5)
    bins_edges = np.arange(0.0, 10.01, 1.0)
    binned_spikes = \
        utils.neuralDataAnalysis.binSpikes(spikes_times=spikes_times,
                                           bins_edges=bins_edges,
                                           time_unit="sec")
    for i in range(len(binned_spikes)):
        assert(binned_spikes[i] == 2)

if __name__ == "__main__":
    test_binSpikes()





import sys
import pdb
import math
import random
import numpy as np
from scipy.io import loadmat
import torch
import pandas as pd
import matplotlib.pyplot as plt
sys.path.append("../../..")
from stats.pointProcess.sampler import Sampler
from stats.pointProcess.tests import KSTestTimeRescalingAnalyticalCorrection, KSTestTimeRescalingNumericalCorrection

def test_compareWilson():
    dataColor = "blue"
    # diagColor = "black"
    cbColor = "red"
    dataLinestyle = "solid"
    # diagLinestyle = "dashed"
    cbLinestyle = "dashed"
    xlabel = "Empirical CDF"
    ylabel = "Model CDF"
    matFilename = "data/dataForKSTestTimeRescalingAnalyticalCorrection.mat"

    loadRes = loadmat(matFilename)
    Y = torch.tensor(loadRes["Y"]).squeeze()
    pk = torch.tensor(loadRes["pk"]).squeeze()
    # empiricalCDF = torch.tensor(loadRes["empiricalCDF"]).squeeze()
    # x = torch.tensor(loadRes["x"]).squeeze()
    # wCB = torch.tensor(loadRes["cb"]).squeeze()
    utSRISIs, uCDF, cb = KSTestTimeRescalingAnalyticalCorrection(Y=Y, pk=pk) # rescaledISIs~exp(\lambda=1.0)
    plt.plot(uCDF, utSRISIs, color=dataColor, linestyle=dataLinestyle)
    # plt.plot([0, 1], [0, 1], color=diagColor, linestyle=diagLinestyle)
    plt.plot([0, 1-cb], [cb, 1], color=cbColor, linestyle=cbLinestyle)
    plt.plot([cb, 1], [0, 1-cb], color=cbColor, linestyle=cbLinestyle)
    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    plt.show()
    pdb.set_trace()

def test_timeRescalingAnalyticalCorrection():
    dt = 1e-3
    t0 = 0.0
    tf = 10.0
    betas = torch.tensor([-3.0, 0.5, -1.0, -.3], dtype=torch.double)
    dataColor = "blue"
    # diagColor = "black"
    cbColor = "red"
    dataLinestyle = "solid"
    # diagLinestyle = "dashed"
    cbLinestyle = "dashed"
    ylabel = "Empirical CDF"
    xlabel = "Model CDF"

    t = torch.arange(t0, tf, dt, dtype=torch.double)
    X = torch.stack((torch.ones(t.shape, dtype=torch.double), 
                     torch.cos(2*math.pi*t), torch.sin(5*math.pi*t), 
                     torch.cos(11*math.pi*t)), 1)
    cifValues = torch.exp(X.matmul(betas))/dt
    sampler = Sampler()
    spikes = torch.tensor(sampler.sampleInhomogeneousPP_timeRescaling(intensityTimes=t, intensityValues=cifValues, T=tf))
    # spikes = torch.tensor(sampler.sampleInhomogeneousPP_thinning(intensityTimes=t, intensityValues=cifValues, T=tf)["inhomogeneous"])
    # spikes = torch.tensor([random.uniform(0, tf) for i in range(len(spikes))])
    bins = np.arange(t0-dt/2, tf+dt/2, dt)
    # start binning spikes using pandas
    cutRes, _ = pd.cut(spikes, bins=bins, retbins=True)
    Y = torch.from_numpy(cutRes.value_counts().values)
    # end binning spikes using pandas
    Y = Y[torch.randperm(Y.shape[0])]
    pk = cifValues*dt
    # pk = cifValues[torch.randperm(cifValues.shape[0])]*dt
    utSRISIs, uCDF, cb = KSTestTimeRescaling(Y=Y, pk=pk) # rescaledISIs~exp(\lambda=1.0)

    plt.subplot(2, 1, 1)
    plt.plot(t, pk)

    plt.subplot(2, 1, 2)
    plt.plot(utSRISIs, uCDF, color=dataColor, linestyle=dataLinestyle)
    # plt.plot([0, 1], [0, 1], color=diagColor, linestyle=diagLinestyle)
    plt.plot([0, 1-cb], [cb, 1], color=cbColor, linestyle=cbLinestyle)
    plt.plot([cb, 1], [0, 1-cb], color=cbColor, linestyle=cbLinestyle)
    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    plt.title("Number of spikes {:d}".format(Y.sum()))

    plt.show()

    pdb.set_trace()

def test_timeRescalingNumericalCorrection():
    dt = 1e-3
    t0 = 0.0
    tf = 10.0
    gamma = 3
    betas = torch.tensor([-3.0, 0.5, -1.0, -.3], dtype=torch.double)

    refColor = "black"
    dataColor = "blue"
    cbColor = "red"
    refLinestyle = "solid"
    dataLinestyle = "solid"
    cbLinestyle = "dashed"
    dataMarker = "*"
    ylabel = "CDF Difference"
    xlabel = "Rescaled Time"

    t = torch.arange(t0, tf, dt, dtype=torch.double)
    X = torch.stack((torch.ones(t.shape, dtype=torch.double),
                     torch.cos(2*math.pi*t), torch.sin(5*math.pi*t),
                     torch.cos(11*math.pi*t)), 1)
    cifValues = torch.exp(X.matmul(betas))/dt
    sampler = Sampler()
    spikes = torch.tensor(sampler.sampleInhomogeneousPP_timeRescaling(intensityTimes=t, intensityValues=cifValues, T=tf))
    # spikes = torch.tensor(sampler.sampleInhomogeneousPP_thinning(intensityTimes=t, intensityValues=cifValues, T=tf)["inhomogeneous"])
    # spikes = torch.tensor([random.uniform(0, tf) for i in range(len(spikes))])
    diffECDFsX, diffECDFsY, cb = KSTestTimeRescalingNumericalCorrection(spikesTimes=spikes, cifTimes=t, cifValues=cifValues, gamma=gamma)

    plt.plot(diffECDFsX, diffECDFsY, color=dataColor, marker=dataMarker, linestyle=dataLinestyle)
    plt.axhline(0, color=refColor, linestyle=refLinestyle)
    plt.axhline(cb, color=cbColor, linestyle=cbLinestyle)
    plt.axhline(-cb, color=cbColor, linestyle=cbLinestyle)
    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    plt.title("Number of spikes {:d}".format(len(spikes)))

    plt.show()

    pdb.set_trace()

if __name__=="__main__":
    # test_timeRescaling()
    test_compareWilson()
    # test_timeRescalingNumericalCorrection()




import sys
import numpy as np
import torch

sys.path.append("../src")
import stats.pointProcess.sampling
import utils.neuralDataAnalysis

def test_thinning(plot=False):
    freq = 1.0 # Hz
    phase = torch.pi/2
    amplitude = 5.0
    nTrials = 100
    bin_size = 0.01
    tol = 100.0
    start_time = 0.0
    end_time = 1.0
    CIF_time_step = 1e-3
    psth_bin_size = 1e-2
    epoch_times = [0.0 for r in range(nTrials)]                                  

    CIF_times = torch.arange(start_time, end_time, CIF_time_step)
    psth_bins_edges = np.arange(start_time,
                               end_time+psth_bin_size,
                               psth_bin_size)
    CIF_values = torch.exp(amplitude*torch.cos(2*torch.pi*freq*CIF_times+phase))
    spikes_times = []
    for r in range(nTrials):
        spikes_times_one_neuron = \
            stats.pointProcess.sampling.sampleInhomogeneousPP_thinning(
                CIF_times=CIF_times, CIF_values=CIF_values)["inhomogeneous"]
        spikes_times.append([np.array(spikes_times_one_neuron)])
    binned_spikes, psth = utils.neuralDataAnalysis.computeBinnedSpikesAndPSTH(
            spikes_times=spikes_times, neuron_index=0,
            trials_indices=np.arange(nTrials),
            bins_edges=psth_bins_edges, time_unit="sec")
    psth_bin_centers = (psth_bins_edges[1:]+psth_bins_edges[:-1])/2
    CIF_values_interp = np.interp(x=psth_bin_centers, xp=CIF_times,
                                  fp=CIF_values)
    mse = np.mean((psth-CIF_values_interp)**2)
    if plot:
        import matplotlib.pyplot as plt
        plt.figure()
        plt.plot(psth_bin_centers, psth, label="PSTH")
        plt.plot(psth_bin_centers, CIF_values_interp, label="CIF")
        plt.legend()
        plt.title("MSE={:.02f}".format(mse))
        plt.xlabel("Time (sec)")
        plt.ylabel("Spike Rate (Hz)")
        import pdb; pdb.set_trace()
    assert(mse<tol)

def test_timeRescaling(plot=False):
    freq = 1.0 # Hz
    phase = torch.pi/2
    amplitude = 5.0
    nTrials = 100
    bin_size = 0.01
    tol = 70.0
    start_time = 0.0
    end_time = 1.0
    CIF_time_step = 1e-3
    psth_bin_size = 1e-2
    epoch_times = [0.0 for r in range(nTrials)]

    CIF_times = torch.arange(start_time, end_time, CIF_time_step)
    psth_bins_edges = np.arange(start_time,
                               end_time+psth_bin_size,
                               psth_bin_size)
    CIF_values = torch.exp(amplitude*torch.cos(2*torch.pi*freq*CIF_times+phase))
    spikes_times = []
    for r in range(nTrials):
        spikes_times_one_neuron = \
            stats.pointProcess.sampling.sampleInhomogeneousPP_timeRescaling(
                CIF_times=CIF_times, CIF_values=CIF_values)
        spikes_times.append([np.array(spikes_times_one_neuron)])
    binned_spikes, psth = utils.neuralDataAnalysis.computeBinnedSpikesAndPSTH(
            spikes_times=spikes_times, neuron_index=0,
            trials_indices=np.arange(nTrials),
            bins_edges=psth_bins_edges, time_unit="sec")
    psth_bin_centers = (psth_bins_edges[1:]+psth_bins_edges[:-1])/2
    CIF_values_interp = np.interp(x=psth_bin_centers, xp=CIF_times,
                                  fp=CIF_values)
    mse = np.mean((psth-CIF_values_interp)**2)
    if plot:
        import matplotlib.pyplot as plt
        plt.figure()
        plt.plot(psth_bin_centers, psth, label="PSTH")
        plt.plot(psth_bin_centers, CIF_values_interp, label="CIF")
        plt.legend()
        plt.title("MSE={:.02f}".format(mse))
        plt.xlabel("Time (sec)")
        plt.ylabel("Spike Rate (Hz)")
        import pdb; pdb.set_trace()
    assert(mse<tol)

if __name__ == "__main__":
    test_thinning(plot=True)
    test_timeRescaling(plot=True)




import pdb
from abc import ABCMeta, abstractmethod
import math
import numpy as  np

class Kernel(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def _k(self, x1, x2, params):
        return

    @abstractmethod
    def _kGrad(self, x1, x2, params, i):
        return

    def buildKSample(self, x1, x2, params):
        K = np.empty((len(x1), len(x2)))
        K[:] = np.nan
        for i in range(len(x1)):
            for j in range(len(x2)):
                value = self._k(x1=x1[i], x2=x2[j], params=params)
                K[i,j]=value
        return K

    def buildKSampleGrad(self, x1, x2, params):
        answer = []
        for l in range(len(params)):
            emptyMatrix = np.empty((len(x1), len(x2)))
            emptyMatrix[:] = np.nan
            answer.append(emptyMatrix)
        for i in range(len(x1)):
            for j in range(len(x2)):
                grad = self._kGrad(x1=x1[i], x2=x2[j], params=params)
                for l in range(len(grad)):
                    answer[l][i,j]=grad[l]
        return answer

class SquaredExponentialKernel(Kernel):

    def _k(self, x1, x2, params):
        l = params[0]
        s2f = params[1]**2
        s2n = params[2]**2
        answer = s2f*math.exp(-1/(2*l**2)*(x2-x1)**2)

        if x1==x2:
            answer = answer + s2n
        return answer

    def _kGrad(self, x1, x2, params):
        l = params[0]
        sf = params[1]
        sn = params[2]

        dl = sf**2/l**3*(x1-x2)**2*math.exp(-1/(2*l**2)*(x1-x2)**2)
        dsf = 2*math.exp(-1/(2*l**2)*(x1-x2)**2)*sf
        dsn = 2*(1 if x1==x2 else 0)*sn
        answer = np.array([dl, dsf, dsn])

        return(answer)

class PeriodicRandomFunctionKernel(Kernel):

    def _k(self, x1, x2, params):
        l = params[0]
        s2f = params[1]**2
        s2n = params[2]**2
        answer = s2f*math.exp(-(2*math.sin((x2-x1)/2)**2)/l**2)

        if x1==x2:
            answer = answer + s2n
        return answer

    def _kGrad(self, x1, x2, params):
        l = params[0]
        sf = params[1]
        sn = params[2]

        dl = sf**2*math.exp(-(2*math.sin((x2-x1)/2)**2)/l**2)*4*math.sin((x2-x1)/2)**2/l**3
        dsf = 2*sf*math.exp(-(2*math.sin((x2-x1)/2)**2)/l**2)
        dsn = 2*(1 if x1==x2 else 0)*sn
        answer = np.array([dl, dsf, dsn])

        return(answer)



import numpy as np


def binSpikes(spikes, bins_edges):
    bin_counts, bins_edges_output = np.histogram(input=spikes,
                                                   bins=bins_edges)
    bin_centers = (bins_edges_output[:-1]+bins_edges_output[1:])/2.0
    return bin_counts, bin_centers


def clipSpikesTimes(spikes_times, from_time, to_time):
    nTrials = len(spikes_times)
    clipped_spikes_times = [[]] * nTrials
    for r in range(nTrials):
        clipped_spikes_times[r] = clipTrialSpikesTimes(
            trial_spikes_times=spikes_times[r],
            from_time=from_time, to_time=to_time)
    return clipped_spikes_times


def clipTrialSpikesTimes(trial_spikes_times, from_time, to_time):
    nNeurons = len(trial_spikes_times)
    clipped_trial_spikes_times = [[]] * nNeurons
    for n in range(nNeurons):
        clipped_trial_spikes_times[n] = clipNeuronSpikesTimes(
            neuron_spikes_times=trial_spikes_times[n],
            from_time=from_time, to_time=to_time)
    return clipped_trial_spikes_times


def clipNeuronSpikesTimes(neuron_spikes_times, from_time, to_time):
    clipped_neuron_spikes_times = neuron_spikes_times[
        np.logical_and(from_time <= neuron_spikes_times,
                          neuron_spikes_times < to_time)]
    return clipped_neuron_spikes_times


def offsetSpikeTimes(spikes_times, offset):
    nTrials = len(spikes_times)
    offsetted_spikes_times = [[]] * nTrials
    for r in range(nTrials):
        offsetted_spikes_times[r] = offsetTrialSpikesTimes(
            trial_spikes_times=spikes_times[r], offset=offset)
    return offsetted_spikes_times


def offsetTrialSpikesTimes(trial_spikes_times, offset):
    nNeurons = len(trial_spikes_times)
    offsetted_trial_spikes_times = [[]] * nNeurons
    for n in range(nNeurons):
        offsetted_trial_spikes_times[n] = trial_spikes_times[n]+offset
    return offsetted_trial_spikes_times


def removeUnits(spikes_times, units_to_remove):
    nTrials = len(spikes_times)
    spikes_times_woUnits = [[]] * nTrials
    for r in range(nTrials):
        spikes_times_woUnits[r] = \
                removeUnitsFromTrial(trial_spikes_times=spikes_times[r],
                                     units_to_remove=units_to_remove)
    return spikes_times_woUnits


def removeUnitsFromTrial(trial_spikes_times, units_to_remove):
    nNeurons = len(trial_spikes_times)
    spikes_times_woUnits = []
    for n in range(nNeurons):
        if n not in units_to_remove:
            spikes_times_woUnits.append(trial_spikes_times[n])
    return spikes_times_woUnits


def selectUnitsWithLessSpikesThanThrInAllTrials(spikes_times, thr):
    nTrials = len(spikes_times)
    nNeurons = len(spikes_times[0])
    selected_units = set([i for i in range(nNeurons)])
    for r in range(nTrials):
        selected_trial_units = selectUnitsWithLessSpikesThanThrInTrial(
            spikes_times=spikes_times[r], thr=thr)
        selected_units = selected_units.intersection(selected_trial_units)
    answer = list(selected_units)
    return answer


def selectUnitsWithLessSpikesThanThrInAnyTrial(spikes_times, thr):
    nTrials = len(spikes_times)
    selected_units = set([])
    for r in range(nTrials):
        selected_trial_units = selectUnitsWithLessSpikesThanThrInTrial(
            spikes_times=spikes_times[r], thr=thr)
        selected_units = selected_units.union(selected_trial_units)
    answer = list(selected_units)
    return answer
    return selected_units


def selectUnitsWithLessSpikesThanThrInTrial(spikes_times, thr):
    nNeurons = len(spikes_times)
    selected_units = set([])
    for n in range(nNeurons):
        if len(spikes_times[n]) < thr:
            selected_units.add(n)
    return selected_units


def removeUnitsWithLessSpikesThanThrInAnyTrial(
        spikes_times, min_nSpikes_perNeuron_perTrial):
    nNeurons = len(spikes_times[0])
    neurons_indices = [n for n in range(nNeurons)]
    units_to_remove = \
        selectUnitsWithLessSpikesThanThrInAllTrials(
            spikes_times=spikes_times,
            thr=min_nSpikes_perNeuron_perTrial)
    spikes_times = removeUnits(spikes_times=spikes_times,
                               units_to_remove=units_to_remove)
    neurons_indices = [n for n in neurons_indices
                       if n not in units_to_remove]
    return spikes_times, neurons_indices


def binSpikes(spikes_times, bins_edges, time_unit):
    # spike_times in msec
    bin_width = bins_edges[1]-bins_edges[0]
    binned_spikes, _ = np.histogram(a=spikes_times, bins=bins_edges)
    binned_spikes = binned_spikes.astype(float)
    if time_unit == "sec":
        binned_spikes *= 1.0/bin_width
    elif time_unit == "msec":
        binned_spikes *= 1000.0/bin_width
    else:
        raise ValueError("time_unit should be sec or msec, but not {}".format(time_unit))
    return binned_spikes


def binMultiTrialSpikes(spikes_times, neuron_index, trials_indices,
                        bins_edges, time_unit):
    mt_binned_spikes = np.empty((len(trials_indices), len(bins_edges)-1),
                                dtype=np.double)
    for i, trial_index in enumerate(trials_indices):
        aligned_spikes_trial_neuron = spikes_times[trial_index][neuron_index]
        binned_spikes = binSpikes(
            spikes_times=aligned_spikes_trial_neuron,
            bins_edges=bins_edges, time_unit=time_unit)
        mt_binned_spikes[i, :] = binned_spikes
    return mt_binned_spikes


def computeBinnedSpikesAndPSTH(spikes_times, neuron_index, trials_indices,
                               bins_edges, time_unit):
    binned_spikes = binMultiTrialSpikes(spikes_times=spikes_times,
                                        neuron_index=neuron_index,
                                        trials_indices=trials_indices,
                                        bins_edges=bins_edges,
                                        time_unit=time_unit)
    psth = np.empty(len(bins_edges)-1, dtype=np.double)
    for j in range(len(bins_edges)-1):
        psth[j] = binned_spikes[:, j].mean()
    return binned_spikes, psth


def computeBinnedSpikesAndPSTHwithCI(spikes_times, neuron_index,
                                     trials_indices, epoch_times,
                                     bins_edges, time_unit,
                                     nResamples, alpha):
    binned_spikes = binMultiTrialSpikes(spikes_times=spikes_times,
                                        neuron_index=neuron_index,
                                        trials_indices=trials_indices,
                                        epoch_times=epoch_times,
                                        bins_edges=bins_edges,
                                        time_unit=time_unit)
    psth = np.empty(len(bins_edges)-1, dtype=np.double)
    psth_ci = np.empty((len(bins_edges)-1, 2), dtype=np.double)
    for j in range(len(bins_edges)-1):
        psth[j] = binned_spikes[:, j].mean()
        bootstrapped_mean = stats.bootstrapTests.bootstrapMean(
            observations=binned_spikes[:, j], nResamples=nResamples)
        psth_ci[j, :] = stats.bootstrapTests.estimatePercentileCI(
            alpha=alpha, bootstrapped_stats=bootstrapped_mean
        )
    return binned_spikes, psth, psth_ci


def alignAndClipSpikeTimes(spike_times, align_times, clip_start_time,
                           clip_end_time):
    nTrials = len(spike_times)
    nNeurons = len(spike_times[0])
    aligned_clipped_spikes_times = []
    for r in range(nTrials):
        aligned_clipped_spikes_times_r = []
        for n in range(nNeurons):
            aligned_spikes_times_rn = spike_times[r][n]-align_times[r]
            aligned_clipped_spikes_times_rn = aligned_spikes_times_rn[
                np.logical_and(clip_start_time <= aligned_spikes_times_rn,
                               aligned_spikes_times_rn < clip_end_time)
            ].tolist()
            aligned_clipped_spikes_times_r.append(
                aligned_clipped_spikes_times_rn)
        aligned_clipped_spikes_times.append(
            aligned_clipped_spikes_times_r)
    return aligned_clipped_spikes_times
